//  $Id: prom_obj.hh,v 1.5 2004/06/16 00:12:25 adams Exp $
//  Author:             Mark F. Adams
// Copyright (c) 2000 by Mark F. Adams 
//  Filename:           prom_obj.hh
//
#ifndef __PROM_OBJECT_H__
#define __PROM_OBJECT_H__

#include "prom_defs.h"

#if defined(__cplusplus)
extern "C"{
#endif
  int prom_intcompar(const void *a, const void *b);
  int prom_intcompar2(const void *a, const void *b);
  int prom_floatcompar(const void *a, const void *b);
#if defined(__cplusplus)
}
#endif

//! class PromCoord: a coodinate 
/*!
  class PromCoord: a coodinate 
*/
class PromCoord {
  friend class PromDTet;
public:  
  PromCoord() { data_[0] = 0.0;data_[1] = 0.0; data_[2] = 0.0; }
  PromCoord(gfloat a,gfloat b,gfloat c){data_[0]=a;data_[1]=b;data_[2]=c;}
  gfloat operator*( const PromCoord &b )const { 
    return data_[0]*b.data_[0] + data_[1]*b.data_[1] + data_[2]*b.data_[2];
  }  
  PromCoord &operator=( const PromCoord &b )  { 
    data_[0] = b.data_[0];
    data_[1] = b.data_[1];
    data_[2] = b.data_[2];
    return *this;   }  
  PromCoord &operator=( const gfloat d )  { 
    data_[0] = d;
    data_[1] = d;
    data_[2] = d;
    return *this;   }  
  PromCoord &operator=( const gfloat *d )  { 
    data_[0] = d[0];
    data_[1] = d[1];
    data_[2] = d[2];
    return *this;   }  
  PromCoord &operator-=( const PromCoord &b )  { 
    data_[0] -= b.data_[0];
    data_[1] -= b.data_[1];
    data_[2] -= b.data_[2];
    return *this;   }
  PromCoord& operator+=( const PromCoord &a)  { 
    data_[0] += a.data_[0];
    data_[1] += a.data_[1];
    data_[2] += a.data_[2];
    return *this;   }
  PromCoord& operator/=( const gfloat d )  {
    data_[0] /= d;
    data_[1] /= d;
    data_[2] /= d;
    return *this;   }
  PromCoord& operator*=( const gfloat d ) {
    data_[0] *= d;
    data_[1] *= d;
    data_[2] *= d;
    return *this; } 
  PromCoord& operator/=( const PromCoord& d )  {
    data_[0] /= d.data_[0];
    data_[1] /= d.data_[1];
    data_[2] /= d.data_[2];
    return *this;   }
  PromCoord& operator*=( const PromCoord& d ) {
    data_[0] *= d.data_[0];
    data_[1] *= d.data_[1];
    data_[2] *= d.data_[2];
    return *this; } 
  PromCoord operator+(const PromCoord a)const{ 
    PromCoord cc;
    cc.data_[0] = data_[0] + a.data_[0];
    cc.data_[1] = data_[1] + a.data_[1];
    cc.data_[2] = data_[2] + a.data_[2];
    return cc;   }    
  PromCoord operator-(const PromCoord a)const { 
    PromCoord cc;
    cc.data_[0] = data_[0]-a.data_[0];
    cc.data_[1] = data_[1]-a.data_[1];
    cc.data_[2] = data_[2]-a.data_[2];
    return cc;   }       
  PromCoord operator*(const gfloat a)const { 
    PromCoord cc;
    cc.data_[0] = data_[0]*a;
    cc.data_[1] = data_[1]*a;
    cc.data_[2] = data_[2]*a;
    return cc;   }          
  PromCoord operator/(const gfloat a)const  { 
    PromCoord cc;
    cc.data_[0] = data_[0]/a;
    cc.data_[1] = data_[1]/a;
    cc.data_[2] = data_[2]/a;
    return cc;   }    
  PromCoord operator-()const  { 
    PromCoord cc;  
    cc.data_[0] = -data_[0];
    cc.data_[1] = -data_[1];
    cc.data_[2] = -data_[2];
    return cc;   }
  gfloat length()const{
    return sqrt((data_[0]*data_[0])+(data_[1]*data_[1])+(data_[2]*data_[2]));
  }  
  gfloat norm2()const{
    return sqrt((data_[0]*data_[0])+(data_[1]*data_[1])+(data_[2]*data_[2]));
  }
  gfloat length2()const {
    return (data_[0]*data_[0]) + (data_[1]*data_[1]) + (data_[2]*data_[2]);
  }
  
  gfloat &operator[]( int ii ){ return data_[ii]; }
  gfloat operator[]( int ii ) const { return data_[ii]; }
  
  // int buffer i/o of PromCoord for MPI 
  int        Write( int **buff )const;
  int        Read( int **buff );
  static int GetWrite_size( int &sz ){ 
    sz = 3*(sizeof(gfloat)/sizeof(int)); 
    return 0; 
  }
  int Archive( FILE *file = stderr, const PromArchiveType type = PROM_PRINT );
private:
  gfloat data_[3];
};

#define MEM_FILE_FLAG (9999)
class PromNode;
class PromMatrix;
class PromComplexMatrix;
class PromOptions;
//! class PromContext: global data
/*!
  class PromContext: global data
*/
class PromContext
{
  // Disable the copy constructor and assignment by default so you will get
  //   compiler errors instead of unexpected behaviour if you pass objects
  //   by value or assign objects.
  //void operator=( const PromContext & objectSrc );     // no implementation
protected:
  PromContext(){} // no real copies of this
public:    
  static int GetNextFactor( const PromOptions &opts, const int np,
			    const int lastf, const int nneq, int *nf );
  static int getNextFactor( const PromOptions &opts, const int np,
			    const int lastf, const int gnodes );
  // util functions 
  static int WriteMat_private(const PromMatrix *, FILE *file, 
			      const int ioff, const int joff, const bool trans);
  static int WriteMat_private_i(const PromComplexMatrix *, FILE *file, 
				const int ioff, const int joff, const bool trs);
  static int WriteMat( const PromMatrix *, char *n, int l );
  static int PrintMatInfo(const PromMatrix *AA, const int level,
			  const int factor, const int mype, const int np );
  
  // array i/o functions
  static int printArr( FILE *file, const int *arr, const int n,
		       const char *header = "%d", const int step = 0 );
  static int printArr( FILE *file, const bool *arr, const int n,
		       const char *header = "%d", const int step = 0 );
  static int printArr( FILE *file, const int *arr, const int n, const int aux,
		       const char *header = "%d", const int step = 0 );
  static int printArr( FILE *file, const double *arr, const int n,const int ax,
		       const char *header = "%d", const int step = 0 );
  static int printArr( FILE *file, const double *arr, const int n,
		       const char *header = "%d", const int step = 0 );
  static int printArr( FILE *file, const float *arr, const int n,
		       const char *header = "%d", const int step = 0 );
  static int scanArr( FILE *file, int **arr, int *n,
		      const char *header = "%d", const int step = 0 );
  static int scanArr( FILE *file, int **arr, int *n, int *aux,
		      const char *header = "%d", const int step = 0 );
  static int scanArr( FILE *file, double **arr, int *n,
		      const char *header = "%d", const int step = 0 );
  static int scanArr( FILE *file, float **arr, int *n,
		      const char *header = "%d", const int step = 0 );
  static int WriteMetis( char *name, const int *xadj, const  int *adjncy, 
			 const int *vwgts, const int *ewgts, PromNode **nodes,
			 const int nNodes, const int numadjac ); 
  // buffer i/o of int for MPI
  static int ScanArrMPI( int myn, int *&array, MPI_Comm Comm );
  static int TInvert( const double src[], double dest[], const int NN,
		      const int LDA = -1 ); 
  static int Write( const int val, int *&buff ){ 
    *buff++ = val; 
    return 0; 
  }
  static int Read( int &val, int *&buff ){ 
    val = *buff++; 
    return 0; 
  }
  static int Read( int &val, int **buff ){ 
    val = **buff; *buff += 1; 
    return 0;
  }
#if !defined(PROM_NO_SYSTEM_BOOL)
  static int Read( bool &val, int *&buff ){ 
    val = (bool)*buff++; 
    return 0; 
  }
  static int Read( bool &val, int **buff ){ 
    val = (bool)**buff; *buff += 1; 
    return 0;
  }
  static int Write( const bool val, int **buff ){
    **buff = (int)val; *buff += 1;
    return 0;
  }
#endif
  static int Write( const int val, int **buff ){
    **buff = val; *buff += 1;
    return 0;
  }
  static int Write( const int num, const int *arr, int *&buff ){ 
    *buff++ = num; 
    for( int i = 0; i < num ; i++ ) *buff++ = arr[i]; 
    return 0;
  } 
  static int Read( int &num, int *arr, int *&buff ){ 
    num = *buff++;
    for(int i = 0 ; i < num ; i++ ) arr[i] = *buff++;
    return 0;
  }

  static int Write( const double val, int *&buff ){
    double *fp = (double*)buff;  *fp++ = val;
    buff = (int*)fp;
    return 0; 
  }
  static int Read( double &val, int *&buff ){ 
    double *fp = (double*)buff; val = *fp++;
    buff = (int*)fp;
    return 0;
  }
  static double **pivot45( const int j , double a[4][5] ){
    double temp, max = fabs(a[j][j]);
    int jj, i;  
    for ( jj = j, i = j + 1 ; i < 4 ; i++ ){
      temp = fabs(a[i][j]);
      if( temp > max ) {
	jj = i;
	max = temp;
      }
    }
    if ( jj == j ) return NULL;
    else {
      // swap rows j & jj    swap5 ( a[jj], a[j]);   
      double *aa = a[jj], *bb = a[j];
      double temp1 = aa[0];  aa[0] = bb[0];  bb[0] = temp1;
      double temp2 = aa[1];  aa[1] = bb[1];  bb[1] = temp2;
      double temp3 = aa[2];  aa[2] = bb[2];  bb[2] = temp3;
      double temp4 = aa[3];  aa[3] = bb[3];  bb[3] = temp4;
      double temp5 = aa[4];  aa[4] = bb[4];  bb[4] = temp5; 
      return (double **)a;
    }
  }
  static gfloat cross( const PromCoord&, const PromCoord&,
		       const PromCoord&_NODE, PromCoord *ret = NULL );
  static gfloat crossDot( const PromCoord&, const PromCoord&,
			  const PromCoord&, const PromCoord&,
			  PromCoord *cross = NULL );
  static int CheckClearMPI( MPI_Comm, const int mype, const int verbose );
  // gloabal data
  static char stateNames_[PROM_NUMSTATE+1][12];
  static char typeNames_[PROM_NUMTOPO+1][12];
  static int  prom_logs_[LOG_TOT+1];
  static int  prom_stages_[11];
};

template <class T> class PromList ;
//! class PromStack (partial class)
/*!
  - This is either a list or an elelemnt of a list.
  - If this is alist then first "next" is the first element of the list.
*/
class PromStack
{
  friend class PromAggGrid;
  friend class PromBank;
  friend class PromParentBank;
  // Disable the copy constructor and assignment by default so you will get
  //   compiler errors instead of unexpected behaviour if you pass objects
  //   by value or assign objects.
  PromStack( const PromStack& objectSrc );              // no implementation
  //void operator=( const PromStack& objectSrc );     // no implementation
protected:  
  PromStack( PromStack *n = NULL ){ next_ = n; }  
  virtual ~PromStack(){ next_ = NULL; }            
public:
  bool   isChildListEmpty() const { return (next_ == NULL); }
  bool   findChild( const PromStack *n ) const ;
  bool   removeChild( PromStack *n );
  int    removeChildren( PromList<PromStack*> * );
  int    ClearChildren();
  int    GetChildren( PromList<PromStack*> * ) const ;
  int    numChildren() const ;
  int    addChild( PromStack *n ){
    assert(this!=NULL); 
    if( n->next_ ) return 1; // only in one list at a time
    n->next_ = next_;
    next_ = n;
    return 0;
  }
  PromStack *PopChild(){ 
    assert(this!=NULL); PromStack *ret = NULL;
    if( next_ ) { ret = next_; next_ = next_->next_; ret->next_ = NULL; } 
    return ret;
  }
  PromStack *getChild() const{assert(this!=NULL); return next_;}
private:
  PromStack *next_;
};

//! class PromGlobalID (meant to be a partial class)
/*!
  - general Global ID for an object < lid, proc >
*/
class PromGlobalID 
{ 
  friend class PromNode;
public:
  PromGlobalID( int i = -9, int p = -9 );
  ~PromGlobalID();
  PromGlobalID &operator=( const PromGlobalID &a );
  void reset() { procID_ = lIndex_ = -9; } 
  int Write( int **ppb )const{ 
    int *pb = *ppb; *pb++ = lIndex_;  *pb++ = procID_; *ppb = pb;
    return 0;
  }
  int Read( int **ppb ){
    int *pb = *ppb; lIndex_ = *pb++;  procID_ = *pb++; *ppb = pb;
    return 0;
  }
  static int GetWrite_size( int &sz ){ sz = 2; return 0;}  
  bool operator==(const PromGlobalID &a)const{
    assert(lIndex_ >= 0); return lIndex_ == a.lIndex_;
  }
  bool operator!=(const PromGlobalID &a)const {return lIndex_ != a.lIndex_;}
  bool operator<(const PromGlobalID &a)const  {return lIndex_ <  a.lIndex_;}
  bool operator>( const PromGlobalID &a)const {return lIndex_ >  a.lIndex_;}
  bool operator<=( const PromGlobalID &a)const{return lIndex_ <= a.lIndex_;}
  bool operator>=( const PromGlobalID &a)const{return lIndex_ >= a.lIndex_;}
  bool operator==( int i ) const { assert(lIndex_ >= 0);  return lIndex_ == i;}
  bool operator!=( int i ) const {   return lIndex_ != i;}
  bool operator<( int i )  const {   return lIndex_ < i;}
  bool operator>( int i )  const {   return lIndex_ > i;}
  bool operator<=( int i ) const {   return lIndex_ <= i;}
  bool operator>=( int i ) const {   return lIndex_ >= i;}
  int operator+( int i ) const { assert(lIndex_ >= 0);    return lIndex_ + i; }
  int operator-( int i ) const { assert(lIndex_ >= 0);    return lIndex_ - i; }
  
  // stored id[ gloabal index, addrsp index index]; sp[0,addrsp]
  void setProc(int f){
    assert(this!=NULL);
    procID_ = f;
  }
  void setProc(int f, int &ierr){ 
    assert(this!=NULL); ierr = !(f >= 0); 
    procID_ = f;
  }
  int  proc(int &ierr)const{ 
    assert(this!=NULL);
    ierr = !(procID_>=0); 
    return procID_;} 
  int proc()const{ 
    assert(this!=NULL); assert(procID_>=0); 
    return procID_;} 
  void setLocal( const int ii ){ lIndex_ = ii; }
  int  local(int &ierr)const{ ierr = !(this && lIndex_ >= 0); return lIndex_;} 
  int  local()const{ assert(this!=NULL && lIndex_ >= 0); return lIndex_;} 
  int  isIerr()const{
    return !(procID_>=0 && lIndex_>=0);
  } 
  int  Archive( FILE *file = stderr, const PromArchiveType type=PROM_PRINT);
private:
  int procID_;   // FEAP (real) index, global solver index.
  int lIndex_;
};

//! class PromLocalID 
/*!
  - local full system ID: < global(PETSc), local >.
  - local ID is default value for object.
*/
class  PromLocalID
{
public:
  PromLocalID(int i = -9){
    assert(this!=NULL);
    gIndex_ = -9;
    lIndex_ = i;
  } 
  void reset() { gIndex_ = lIndex_ = -9; } 
  PromLocalID( const PromLocalID &a ){ 
    assert(this!=NULL);
    //fIndex = a.fIndex;
    gIndex_ = a.gIndex_;
    lIndex_ = a.lIndex_;
  }
  PromLocalID &operator=( const PromLocalID &a ){
    assert(this!=NULL);
    lIndex_ = a.lIndex_;
    gIndex_ = a.gIndex_;
    return *this;
  }
  int Write( int **ppb )const{ 
    int *pb = *ppb; *pb++ = lIndex_; *pb++ = gIndex_; *ppb = pb;
    return 0;
  }
  int Read( int **ppb ){
    int *pb = *ppb; lIndex_ = *pb++; gIndex_ = *pb++; *ppb = pb;
    return 0;
  }
  static int GetWrite_size( int &sz ){ sz = 2; return 0; }  

  /*bool operator==( const PromLocalID &a )const { return lIndex_ == a.lIndex_; }
    bool operator!=( const PromLocalID &a )const { return lIndex_ != a.lIndex_; }
    bool operator<( const PromLocalID &a )const  { return lIndex_ < a.lIndex_; }
    bool operator>( const PromLocalID &a )const  { return lIndex_ > a.lIndex_; }
    bool operator<=( const PromLocalID &a )const { return lIndex_ <= a.lIndex_; }
    bool operator>=( const PromLocalID &a )const { return lIndex_ >= a.lIndex_; }
    bool operator==( int i )const { return lIndex_ == i; }
    bool operator!=( int i )const { return lIndex_ != i; }
    bool operator<( int i )const { return lIndex_ < i; }
    bool operator>( int i )const { return lIndex_ > i; }
    bool operator<=( int i )const { return lIndex_ <= i; }
    bool operator>=( int i )const { return lIndex_ >= i; }
    int operator+( int i )const { return lIndex_ + i; }
    int operator-( int i )const { return lIndex_ - i; }*/
    
  // stored id[ gloabal index, addrsp index index]; sp[0,addrsp]
  int  global(int &ierr)const{ierr = !(this && gIndex_ >= 0);return gIndex_;}
  int  global()const{ assert(this!=NULL && gIndex_ >= 0); return gIndex_;}
  void setGlobal( int g ){ assert(this!=NULL); gIndex_ = g; }
  int  operator()(int &ierr)const{ierr=!(this && lIndex_ >= 0);return lIndex_;}
  int  operator()()const{ assert(this!=NULL && lIndex_ >= 0); return lIndex_;}
  PromLocalID &operator=( int i ){ lIndex_ = i; return *this; }

  int  isIerr()const{return !(gIndex_ >= 0 && lIndex_ >= 0 );}
  int  Archive( FILE *file = stderr, const PromArchiveType type = PROM_PRINT );
private:
  int gIndex_;
  int lIndex_;   // FEAP (real) index, global solver index.
};

#endif // __PROM_OBJECT_H__
