//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: geo_coar_tri.C,v 1.2 2004/03/01 16:15:18 adams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           geo_coar_tri.C
//
#include "prom_grid.hh"

static PromTable<PromNode*> *hull = NULL;
static PromTable<PromDTet*> *list1 = NULL, *oldElems = NULL, *list2 = NULL; 
#define RTOL 1.e-12
  
#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::Triangulate" 
/* PromGeoGrid::Triangulate ************************************************
 *
 *  - make tetrahedra with Watson's algo, nodes have approx adjac.
 * 
 *  INPUT:
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code 
 */
int PromGeoGrid::Triangulate()
{ 
  TablePos     tpos;
  PromGeoNode *root,*curr, *n1;
  PromDTet    *newelem, *elem;                          
  int          i,j,ierr,xx,yy;
  bool         discon, ok;
  const int    todo = nNodes_,mype = comm_.mype(), verb = options_.verbose_>1;

  // INITIAL elementss (12) of env nodes AND the FIRST NODE
  static const int boxID[12][3] = { {0,6,4},{0,2,6},{0,4,5},{0,5,1},
                                    {0,1,2},{1,3,2},{1,5,3},{3,5,7},
                                    {3,7,6},{3,6,2},{5,4,6},{5,6,7} };
  if(verb>1)PetscPrintf(PETSC_COMM_SELF,"[%d]%s start\n",
				      mype,__FUNCT__);
  if( todo < 1 ) {
    PetscPrintf(PETSC_COMM_SELF,"[%d]%s !enough node!\n",mype,__FUNCT__,todo);
    return 0;
  }
  if( !list1 ) {
    list1 = new PromTable<PromDTet*>(147);    CHKNEW_void(list1,1);
    list2 = new PromTable<PromDTet*>(147);    CHKNEW_void(list2,1);
    oldElems = new PromTable<PromDTet*>(153); CHKNEW_void(oldElems,1);
    hull = new PromTable<PromNode*>(183);     CHKNEW_void(hull,1);
  }

  PromList<PromGeoNode*> nextRoots(100);  gfloat temp, tt;
  PromTable<PromGeoNode*> doneNodes( (todo*3)/2 + 5 );  
  PromCoord maxCoord(-1.e9,-1.e9,-1.e9), minCoord(1.e9,1.e9,1.e9);
  DeleteTets(); 
  // get root, set box  
  for( xx = 0, yy = 0, root = NULL; xx < nNodes_ ; xx++ ){
    curr = dynamic_cast<PromGeoNode*>(nodes_[xx]); assert(curr!=NULL);
    curr->mark(FALSE); assert(curr->symbData_->elemList.isEmpty());
    if( !root && curr->symbData_->adjac.getCount() >= yy ) {
      root = curr; yy = curr->symbData_->adjac.getCount();
    }
    // collect envelope 
    for( j = 0 ; j < 3 ; j++ ) {
      gfloat temp = curr->symbGeoData_->coord[j];
      if ( temp > maxCoord[j] ) maxCoord[j] = temp;
      if ( temp < minCoord[j] ) minCoord[j] = temp;
    }
  }
  // expand envelope
  if( !root ){
    if(verb>1) PetscPrintf(PETSC_COMM_SELF,"[%d]%s no root\n",
					 mype,__FUNCT__,todo);
    assert(nLocalNd_ == 0);
    return 0;
  }
  for( j = 0, tt = 0.0 ; j < 3 ; j++ ) {
    temp = maxCoord[j] - minCoord[j]; assert(temp >= 0.0);
    maxCoord[j] = temp; tt += temp;
  }
  tt /= 100.0; // some length
  if( tt == 0.0 ) tt = 1.0; // really degenerate case (one node)
  const gfloat scale = 15.0;      // maxCoord has length of box now!
  for( i = 0 ; i < 4 ; i++ ) {
    env_[i]->symbGeoData_->coord[2]   = minCoord[2] - (scale*maxCoord[2] + tt);
    env_[i+4]->symbGeoData_->coord[2] = minCoord[2] + maxCoord[2] + 
      (scale*maxCoord[2] + tt);
  }
  for( i = 0 ; i < PROM_NUMENV ; i += 2 ) {
    env_[i+0]->symbGeoData_->coord[0] = minCoord[0] - (scale*maxCoord[0] + tt);
    env_[i+1]->symbGeoData_->coord[0] = minCoord[0] + maxCoord[0] + 
      (scale*maxCoord[0] + tt);
  }
  for( i = 0 ; i < PROM_NUMENV ; i += 4 ) {
    temp = minCoord[1] - (scale*maxCoord[1]+tt);
    env_[i+0]->symbGeoData_->coord[1] = env_[i+1]->symbGeoData_->coord[1]=temp;
    temp = minCoord[1] + maxCoord[1] + (scale * maxCoord[1] + tt); 
    env_[i+2]->symbGeoData_->coord[1] = env_[i+3]->symbGeoData_->coord[1]=temp;
  }
  // make box 
  for ( i = 0 ; i < 12 ; i++ ) {
    newelem = newTet(env_[boxID[i][0]],env_[boxID[i][1]],env_[boxID[i][2]],
		     root);    
    elemList_.addTail( newelem  ); // add to list allways
    newelem->AddSelfToNodes();              // add to node's elem list
    assert( newelem->neigh_[3] == NULL );
  }
  for(i=0;i<8;i++)assert(env_[i]->isEnvelope());
  // add root
  doneNodes.add(root);  assert(!root->isMarked());  root->mark();  
  int count = 1;        // num done
  root->SetElemAdjac(); // set elemensts 'adj' - exterior will stay  null
  if( todo < 200 ) {
    for( xx = 1 ; xx < nNodes_ ; xx++ ){
      curr = dynamic_cast<PromGeoNode*>(nodes_[xx]);     assert(curr!=NULL);
      if( curr == root ) continue;
      elem = curr->setElemWithList( &root->symbData_->elemList,parBank_, 
				    mype, 100, RTOL, FALSE, TRUE );
      assert(curr!=NULL);   assert(!curr->isMarked());
#if defined(PETSC_USE_BOPT_g)
      //{double t;assert(elem->isInCirc(curr->symbGeoData_->coord,&t));}
#endif
      ierr = AddToMesh( curr, elem );                CHKERRQ(ierr);
      doneNodes.add(curr); assert(!curr->isMarked()); curr->mark(); count++; 
    }
    assert(todo==count);
  } 
  // insert nodes to the mesh  - skip done nodes
  int timer = 0;             // clean elem list
  while(count != todo){ 
    while( root ) {
      assert( root->isMarked() ); assert(!root->symbData_->elemList.isEmpty());
      tpos = root->symbData_->adjac.getHeadPosition(); //assert(tpos!=0);
      while( tpos ) {
	n1 = dynamic_cast<PromGeoNode*>(root->symbData_->adjac.getNext(tpos));
	// skip done nodes and just undone nodes
	if( doneNodes.find( n1 ) ) continue;
	assert( n1->symbData_->elemList.isEmpty() ); 
	// get a starter elem  
	elem = n1->setElemWithList( &root->symbData_->elemList,parBank_, 
				    mype, 10, RTOL, FALSE, TRUE );
	if( elem == NULL ) continue;
	// PetscPrintf(PETSC_COMM_SELF,"[%d]%s() not found\n",mype,__FUNCT__);
	assert(!elem->isDead());
#if defined(PETSC_USE_BOPT_g)
	//{double t;assert(elem->isInCirc(n1->symbGeoData_->coord,&t));}
#endif
	ierr = AddToMesh( n1, elem );               CHKERRQ(ierr);
	doneNodes.add(n1); assert(!n1->isMarked()); n1->mark(); count++;
	nextRoots.addTail( n1 );
	// maintain mem.
	if ( timer++ > 200 ) { //timer
	  CleanElemList();      
	  timer = 0;
	} 
      }
      if( !nextRoots.isEmpty() )  {
	do root = nextRoots.removeHead();
	while( !root->isMarked() && !nextRoots.isEmpty() );
	if( !root->isMarked() ) {
	  assert( root->symbData_->elemList.isEmpty() ); 
	  root = NULL;
	}
	else assert( !root->symbData_->elemList.isEmpty() ); 
      }
      else root = NULL;
    }
    discon = (doneNodes.getCount() != todo);
    if( verb>1 || (mype == 0 && discon) ) 
      PetscPrintf( PETSC_COMM_SELF,
		   "[%d]%s:%s: added %d/%d nodes (%d/%d)\n",
		   mype, __FUNCT__, discon ? "WARNING - disconnected" : "", 
		   count, todo, doneNodes.getCount(), todo);
    assert(doneNodes.getCount() <= todo);
    if( !discon ) break;
    
    // add next root
    for( root = NULL, xx = 0 ; xx < nNodes_  && root == NULL; xx++ ){
      curr = dynamic_cast<PromGeoNode*>(nodes_[xx]); assert(curr!=NULL);
      if( curr->isMarked() ) continue; 
      assert( curr->symbData_->elemList.isEmpty() ); 
      if( curr->symbData_->adjac.isEmpty() && verb>1 ) {
	PetscPrintf(PETSC_COMM_SELF,"[%d]%s WARN: %d no adjs,isg=%d,state=%s\n",
		    mype,__FUNCT__,xx,curr->isGhost(), 
		    PromContext::stateNames_[curr->getState()]);
      }
      root = curr; // next root
      // add it - get elem
      if(verb>1) PetscPrintf(PETSC_COMM_SELF,
			     "[%d]%s WARN: %d EXPEN$IVE,isg=%d,state=%s\n",
			     mype,__FUNCT__,xx,curr->isGhost(), 
			     PromContext::stateNames_[curr->getState()]);
      elem = NULL;
      Position pos2, pos = elemList_.getHeadPosition(); assert(pos!=NULL);
      while( (pos2 = pos) && !elem ) {	
	PromPoly *poly = elemList_.getNext(pos);
	PromDTet* e = dynamic_cast<PromDTet*>(poly); assert(e!=NULL);
	shpfloat fb[4];
	if( e->isDead() ) elemList_.removeAt(pos2);
	else if( e->isIn(root->symbGeoData_->coord,fb,RTOL,mype,&ok)){
	  elem = e;
	}
      }
      if( !elem ) {
	assert(curr!=NULL); root = NULL; // no good
	SETERRQ(1,"no elem!!!!");
      }
    }
    if( elem && root ){
      assert(root->symbData_->elemList.isEmpty()); 
#if defined(PETSC_USE_BOPT_g)
      {double t;assert(elem->isInCirc(root->symbGeoData_->coord,&t));}
#endif
      ierr = AddToMesh( root, elem );               CHKERRQ(ierr);
      doneNodes.add(root); assert(!root->isMarked()); root->mark(); count++;
      assert( !root->symbData_->elemList.isEmpty() );
    }
    else{ assert(doneNodes.getCount() == todo); break; }
  }
  assert(count == todo);
  //  
  for( xx = 0 ; xx < nNodes_ ; xx++ ){
    curr = dynamic_cast<PromGeoNode*>(nodes_[xx]);
    curr->mark( FALSE ); assert( curr->isChildListEmpty() );
  }
  // clean up buffers
  delete list1; delete list2; delete  oldElems; delete hull;
  list1 = NULL; list2 = NULL; oldElems = NULL; hull = NULL;
  CleanElemList();   

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::AddToMesh" 
/* PromGeoGrid::AddToMesh *************************************************
 *
 *  add node to a mesh (Watson's insertion algo)
 *
 *  INPUT:
 *    - node: node to add
 *    - elem: element to start
 *    - count: number of done nodes (in/out)
 *
 *  SIDE EFFECTS:
 *    - removes old elements from many nodes 'ElemList'
 *    - creates new elems
 *    - updates nodes 'elemList'
 *    - sets elems 'adj' pointers
 *
 *  RETURN:
 *    - PET cerror code 
 */
int PromGeoGrid::AddToMesh( PromGeoNode *node, PromDTet *elem0 )
{
  int          ierr;            assert(node!=NULL); assert(!node->isMarked());
  PromGeoNode *n1;       assert(elem0 != NULL && !elem0->isDead());
  PromDTet    *elem;
  gfloat       v2 = 0.;
  Position     pos, pos2;
   
  // make cavity
  if( !node->symbData_->elemList.isEmpty() ) 
    SETERRQ(1,"!node->symbData_->elemList.isEmpty()");
  oldElems->removeAll();
  ierr = GetCavity( elem0, node, oldElems, &v2);          CHKERRQ(ierr);
  assert( !oldElems->isEmpty() );

  // make list of hull nodes
  hull->removeAll();
  pos = node->symbData_->elemList.getHeadPosition();
  while ( pos ) { 
    elem = dynamic_cast<PromDTet*>(node->symbData_->elemList.getNext(pos));
    assert(elem!=NULL);
    // add ext. nodes to hull
    for ( int j = 0 ; j < 3 ; j++ ) hull->add(elem->nodes_[j]);
    assert( elem->nodes_[3] == node);
  }  

  // add new elems to hull nodes 'elemList', add new elems to structures list
  pos = node->symbData_->elemList.getHeadPosition();
  while( pos ) { 
    elem = dynamic_cast<PromDTet*>(node->symbData_->elemList.getNext(pos));
    assert(elem!=NULL && elem->isMarked() && !elem->isDead());
    elem->mark( FALSE ); // unmark, as cavity is completed
    elemList_.addTail( elem ); // add to list allways
    
    assert( elem->nodes_[3] == node );
    for( int j = 0 ; j < 3 ; j++ ) {
      n1 = dynamic_cast<PromGeoNode*>(elem->nodes_[j]);
      n1->symbData_->elemList.addTail( elem );     //assert( n1->isMarked() );
    }
  }

  // deallocate old tets
  TablePos tpos = oldElems->getHeadPosition();            assert(tpos!=0);
  while( tpos ){
    elem = oldElems->getNext(tpos);  
    assert(elem->isDead()); elem->kill(0); // temp 'kill'
    Kill( elem );  // this element is being consummed by IP
    assert( elem->isDead() );
  }
  // remove old elements from hull nodes 'elemList' - clear hull list
  while( !hull->isEmpty() ) {
    PromNode *curr = hull->removeHead();  assert( curr != node ); 
    // remove old elems
    pos = curr->symbData_->elemList.getHeadPosition();
    while ( (pos2 = pos) != NULL ) {
      elem = dynamic_cast<PromDTet*>(curr->symbData_->elemList.getNext(pos));
      assert(elem!=NULL);
      if ( elem->isDead() ) { 
        curr->symbData_->elemList.removeAt( pos2 );
	assert(!curr->symbData_->elemList.isEmpty() );
      }
    }
  }

  // set adjacentcies for thee elems - exterior done in AddToMesh
  node->SetElemAdjac();

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::GetCavity" 
/* PromGeoGrid::GetCavity ***********************************************
 *
 *   - get Delaunay cavity (Watsons's insertion algorithm) of 'node'
 *
 *   INPUT:
 *     - thisEl: element that the coord is in
 *     - node: node to search for.
 *     - oldEles: return list of old (dead) elems
 *     - vol: current volume (output)
 *
 *   SIDE EFFECTS:
 *     - makes new elems (with adjac[3] == elem out of cavity for a face)
 *     - new elems are in 'node->symbData_->elemList'
 *     - add 'thisEl' to 'oldEles'
 *
 *   RETURN:
 *     - PET cerror code 
 */
int PromGeoGrid::GetCavity( PromDTet *thisEl, PromGeoNode *node, 
			    PromTable<PromDTet*> *oldEles, gfloat *vol)
{
  int mype = comm_.mype();
  PromDTet      *testElem, *newel, *currEl;
  TablePos       tpos1; double t2,t3;
  while( !list1->isEmpty() ) list1->removeHead(); // preserve corpses
  while( !list2->isEmpty() ) list2->removeHead(); 
  assert( oldEles->isEmpty() );  

  // breadth first search for all of cavity
  list1->add( thisEl );                            assert( !thisEl->isDead() );
  tpos1 = list1->getHeadPosition();            assert(tpos1!=0);
  do {
    currEl = list1->getNext( tpos1 );  assert(!currEl->isDead());
    oldEles->add( currEl );  currEl->kill(); // temp. mark
    // test adjacent elems, for boundry
    for( int ii = 0 ; ii < 4 ; ii++ ) {
      testElem = dynamic_cast<PromDTet*>(currEl->neigh_[ii]); // could be null
      assert(testElem != NULL || currEl->neigh_[ii] == NULL);
      if( testElem && (list1->find(testElem) || list2->find(testElem) || 
		       oldEles->find(testElem)) ) continue;
      assert( testElem==NULL || !testElem->isDead() );
      // see if face is still alive of if bound
      t2 = -1.;
      bool eatit = FALSE;
      if( testElem ){ 
	assert(!testElem->isDead());
	eatit = testElem->isInCirc( node->symbGeoData_->coord, &t2 );
	if( !eatit && (t3=testElem->getVolume()) <= 0.0 ) {
	  if( options_.verbose_>1 ){
	    PetscPrintf(PETSC_COMM_SELF,
			"\t\t[%d]***killing zero vol elem. isincirc=%e vol=%e\n",
			mype,t2,t3);
	  }
	  eatit = TRUE; 
	}
      }
      // check for OK volume
      if( eatit ) { // a live face
	assert( testElem != NULL ); list2->add( testElem );     // keep going
      }
      else {  
	// set local face nodes  
	PromGeoNode *n1 = dynamic_cast<PromGeoNode*>(currEl->nodes_[PromPoly::tetID_[ii][0]]);
	PromGeoNode *n2 = dynamic_cast<PromGeoNode*>(currEl->nodes_[PromPoly::tetID_[ii][1]]);
	PromGeoNode *n3 = dynamic_cast<PromGeoNode*>(currEl->nodes_[PromPoly::tetID_[ii][2]]);
	if( n1 == node || n2 == node || n3 == node ) {
	  SETERRQ(1,"bad element, failed to recover from fix below");
	}
	// create and check for sliver (again)
	newel = newTet( n1, n2, n3, node ); // NEW ELEM
	gfloat tvol = newel->getVolume(); int ierr;
	if( tvol < 0.0 || (options_.verbose_ > 1 && tvol == 0.0) ) { 
	  PetscPrintf(PETSC_COMM_SELF,"\n[%d]vol=%le, t2=%le, currEl=%p thisEl=%p testElem=%p\n 1) %le,%le,%le isenv=%d isgh=%d gid=%d \n 2) %le,%le,%le isenv=%d isgh=%d gid=%d \n 3) %le,%le,%le isenv=%d isgh=%d gid=%d\n 4) %le,%le,%le isenv=%d isgh=%d gid=%d\n\n",
		      mype,tvol,t2,currEl,thisEl,testElem,n1->symbGeoData_->coord[0],
		      n1->symbGeoData_->coord[1],n1->symbGeoData_->coord[2],
		      n1->isEnvelope(),n1->isGhost(),n1->index_.global(ierr),
		      n2->symbGeoData_->coord[0],n2->symbGeoData_->coord[1],
		      n2->symbGeoData_->coord[2],n2->isEnvelope(),n2->isGhost(),
		      n2->index_.global(ierr),n3->symbGeoData_->coord[0],
		      n3->symbGeoData_->coord[1],n3->symbGeoData_->coord[2],
		      n3->isEnvelope(),n3->isGhost(),n3->index_.global(ierr),
		      node->symbGeoData_->coord[0],node->symbGeoData_->coord[1],
		      node->symbGeoData_->coord[2],node->isEnvelope(),
		      node->isGhost(),node->index_.global(ierr));
	}
	if( tvol < 0.0 ) { 
	  SETERRQ1(1,"vol=%e, 'predicates.c' compiled with high optimization or 'exactinit()' not called...",tvol);
	}
	else {
	  // make a new elem (found a boundry of cavity)
	  node->symbData_->elemList.addTail( newel ); 
	  newel->mark();                  // mark as new here
	  *vol += newel->getVolume();     // accume volume to check        
	  // if an adjacent elem to cavity - set its 'adj'
	  newel->neigh_[3] = testElem;    // exterior adj - could be null  
	  if( testElem != NULL ) {
	    assert ( !testElem->isDead() );
	    int iii = testElem->SetNeigh( newel, n1, n2, n3 );
	    assert( iii >= 0 && iii < 4 );
	    testElem->neigh_[iii] = newel;
	  }
	} 
      }
    } 
    // move up one level
    if ( !tpos1 ) {
      list1->removeAll();
      while ( !list2->isEmpty() ) {
        currEl = list2->removeHead(); assert( !currEl->isDead() ); 
        list1->add( currEl );
      }
      tpos1 = list1->getHeadPosition();
    }    
  }while( tpos1 ); // while more nodes todo
  
  return 0;
}

#define NTOL_OFFSET 4
#undef __FUNCT__
#define __FUNCT__ "PromGeoGrid::OffsetParts" 
/* int PromGeoGrid::OffsetParts() ****************************************
 *
 *  - seperate unconnected nodes on here (reaches down to move nodes): 
 *      e.g. contact 
 * 
 *  INPUT:
 *    - factor: multiple of max to m
 *    - minCoord: (in or out)
 *    - maxCoord: (in or out)
 *    - coordsIN: calculate bounding box (minCoord,maxCoord)
 *
 *  SIDE EFFECTS:
 *
 *  RETURN:
 *    - PETSc error code  
 */
int PromGeoGrid::OffsetParts( PromGeoGrid * const lastGrid,
			      const PromCoord &factor, PromCoord &minCoord, 
			      PromCoord &maxCoord, const bool coordsIN )
{
  int              xx, nlast_done,color,mype=comm_.mype();  
  bool             coordsDone = FALSE;
  TablePos         tpos;
  PromList<PromGeoNode*> nextRoots(100);  
  PromTable<PromNode*> doneNodes( (nNodes_*3)/2 );  
  PromGeoNode     *root,*curr;

  if(options_.verbose_>2)PetscPrintf(PETSC_COMM_SELF,"[%d]%s start\n",mype,__FUNCT__);
  if( !nodes_ ) return 0;
  // clear
  for( xx = 0 ; xx < nNodes_; xx++ ) nodes_[xx]->mark(FALSE); 

  // set child lists
  const int nnodes = nNodes_, nnodes_last = lastGrid->nNodes_;
  for( xx = 0 ; xx < nnodes_last ; xx++ ) {
    PromNode *curr, *fine = lastGrid->nodes_[xx]; 
    if( fine->getState() == PROM_SELECTED ) {
      curr = fine->symbData_->parent; assert(curr!=NULL); 
    }
    else if( fine->getState() == PROM_DELETED ) {
      curr = fine->symbData_->parent->symbData_->parent; assert(curr!=NULL); 
    }
    else continue; // ???
    curr->addChild( fine );
  }
  //
  root = dynamic_cast<PromGeoNode*>(nodes_[0]); assert(root!=NULL);
  // find sperate bodies and seperate
  color = 1;
  root->mark(); doneNodes.add( root ); root->setColor( color );
  nlast_done = 0; 
  do{ // for each body
    while( root ) { // for all connected nodes
      PromTable<PromNode*> *adj = &root->symbData_->adjac;
      tpos = adj->getHeadPosition(); 
      if( !tpos )SETERRQ(1,"ERROR: too few nodes?, use fewer levels");
      while( tpos ) {
	curr = dynamic_cast<PromGeoNode*>(adj->getNext(tpos));
	if( !curr->isMarked() ) {
	  int add = doneNodes.add(curr); curr->mark(); assert(add);
	  nextRoots.addTail( curr );
	  curr->setColor( color );
	}
      }
      // look for next root
      if( !nextRoots.isEmpty() ) root = nextRoots.removeHead();
      else root = NULL;
    }

    xx = doneNodes.getCount();
    if( nnodes - xx > NTOL_OFFSET && xx > NTOL_OFFSET ) {
      // find env, add stranded (dummy) nodes to done list 
      if( !coordsIN && !coordsDone ) {
	coordsDone = TRUE;
	// coords are OUTput
	maxCoord[0] = maxCoord[1] = maxCoord[2] = -1.e9;
	minCoord[0] = minCoord[1] = minCoord[2] =  1.e9;
	for( xx = 0 ; xx < nNodes_; xx++ ) {
	  curr = dynamic_cast<PromGeoNode*>(nodes_[xx]); 
	  if( curr->symbData_->adjac.isEmpty() ){ 
	    doneNodes.add( curr ); curr->mark(); 
	    curr->setColor( color ); 
	  }
	  else{
	    // collect envelope
	    for( int j = 0 ; j < 3 ; j++ ) {
	      gfloat temp = curr->symbGeoData_->coord[j];
	      if ( temp > maxCoord[j] ) maxCoord[j] = temp;
	      if ( temp < minCoord[j] ) minCoord[j] = temp;
	    }
	  }
	}
      }
      else if( !coordsDone ) {
	coordsDone = TRUE;
	// add nodes w/o adjacs
	for( xx = 0 ; xx < nNodes_ ; xx++ ) {
	  curr = dynamic_cast<PromGeoNode*>(nodes_[xx]); 
	  if( curr->symbData_->adjac.isEmpty() ){
	    doneNodes.add(curr); curr->mark();
	  }
	}
      }
      if( nnodes - doneNodes.getCount() <= NTOL_OFFSET ) {
	break; // recheck
      }
      // offset - floaters and get new root
      PromCoord off1( maxCoord - minCoord );  
      double len = off1.length(); // determinate 
      PromCoord off(len*((color%2==0)?1.:-1.),len*((color%3==0)?1.:-1.),len*((color%2==1)?1.:-1.));  
      off = off * factor;
      for( xx = 0, root = NULL ; xx < nNodes_ ; xx++ ) {
	curr = dynamic_cast<PromGeoNode*>(nodes_[xx]); 
	if( !curr->isMarked() ) { // offset all not done nodes
	  if( !root ) root = curr;
	  curr->symbGeoData_->coord += off;
	  // 
	  PromGeoNode *gc = dynamic_cast<PromGeoNode*>(curr->getChild());
	  while( gc != NULL ) {
	    gc->symbGeoData_->coord += off;
	    gc = dynamic_cast<PromGeoNode*>(gc->getChild());
	    nlast_done++;
	  }
	}
      }
      if( options_.verbose_ > 1 && !coordsIN )
	PetscPrintf( options_.verbose_  == 2 ? PETSC_COMM_WORLD : PETSC_COMM_SELF,
		     "[%d]%s DISCONNECTED (%d/%d) + %e last grid (%d/%d)\n",
		     mype,__FUNCT__,doneNodes.getCount(),nnodes,
		     off.length(),nlast_done,nnodes_last);
      if(!root)SETERRQ(1,"!root");
      if(root->isMarked()) SETERRQ(1,"root->isMarked()"); 
    }
    else{
      // set colors and break
      for( xx = 0 ; xx < nNodes_; xx++ ) {
	curr = dynamic_cast<PromGeoNode*>(nodes_[xx]); 
	if( curr->color() <= 0 ) {
	  curr->setColor( color ); curr->mark(); 
	  int add = doneNodes.add( curr );  assert(add);
	}
      }
      break;
    }
    color++;
    root->mark(); doneNodes.add( root );
    root->setColor( color );
  }while( nnodes > doneNodes.getCount() );   
  assert(nnodes == doneNodes.getCount());

  // unmark
  for( xx = 0 ; xx < nNodes_ ; xx++ ) {
    curr = dynamic_cast<PromGeoNode*>(nodes_[xx]); curr->mark(FALSE);
    curr->ClearChildren();   // clear child list 
  }

  return 0;
}
